import re
from copy import deepcopy
import time
import requests


class DataIter(object):
    """ change or replace data from string, list or dict
    """

    def __init__(self, *rules, data=None, replace_key=False, mode='replace', **kwargs):
        self._rules = rules if data else rules[:-1]
        self._data = deepcopy(data) if data else rules[-1]

        self.replace_key = replace_key
        self.mode = mode
        self.kwargs = kwargs

        assert mode in ['replace', 'update', 'search', 'delete'], f'Invalid mode ... {mode}'

        if self.mode in ['search', 'delete']:
            self._rule_keys = self.get_rule_keys()
        elif self.mode == 'update':
            self._update_dict = self.rules_to_dict()

        self._result = {}

    @property
    def result(self):
        if self.mode in ['replace', 'update', 'delete']: self._result = self.iter_data(self._data)
        if self.mode in ['search']: self.iter_data(self._data)
        if isinstance(self._result, dict):
            return self._result if len(self._result.keys()) != 1 else list(self._result.values())[0]
        else:
            return self._result

    @property
    def rules(self):
        return self._rules

    @rules.setter
    def rules(self, rules):
        self._rules = rules
        self._rule_keys = self.get_rule_keys()
        self._update_dict = self.rules_to_dict()

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = deepcopy(data)

    # --------------------------------------------------------------------------------- core function
    def iter_data(self, data=None):
        """ 遍历整个数据 for search mode and delete mode

        :param data: dict or list or str
        :return: dict or list
        """
        if isinstance(data, dict):
            # process dict data here
            _data = {}
            for k, value in data.items():
                if k == value in [None, '', 'None']: continue
                # process key and value in here which value type is list or dict

                # delete key:value which key in rule key list
                if self.mode == 'delete' and k in self._rule_keys: continue

                trans = False
                if isinstance(value, tuple): value, trans = list(value), True
                value = self._bypass(key=k, value=value, passing='dict')
                if trans: value = tuple(value)

                if isinstance(value, tuple):
                    k, value = value if not trans else (k, value)
                elif isinstance(value, list):
                    k, value = k, [_[1] if isinstance(_, tuple) else _ for _ in value]

                if self.mode == 'replace' and self.replace_key and isinstance(value, (tuple, list, dict)):
                    k = self.replacer(value=k)
                _data[k] = value

        elif isinstance(data, list):
            # process list data here
            _data = []
            for dt in data:
                _data.append(self._bypass(value=dt, passing='list'))

        elif isinstance(data, str):
            _data = self._bypass(value=data, passing='str')
        else:
            raise Exception('Data Type Error ... data type is unsupported')

        return _data

    def _bypass(self, key=None, value=None, passing=None):
        # all key will pass here
        if self.mode == 'search' and key in self._rule_keys: self.update_result(key=key, value=value)
        if self.mode == 'update' and key in self._update_dict.keys(): return key, self._update_dict.get(key)

        if isinstance(value, (list, dict)):
            return self.iter_data(value)
        elif passing == 'dict':
            return self._process_dict_value(key=key, value=value)
        elif passing == 'list':
            return self._process_list_value(value)
        else:
            return self._process_string(value)

    def _process_dict_value(self, key, value):
        # process value in dict here, value is not list or dict
        return self.replacer(key=key, value=value) if self.mode == 'replace' else (key, value)

    def _process_list_value(self, value):
        # process value in list here, value is not list or dict
        return self.replacer(value=value) if self.mode == 'replace' else value

    def _process_string(self, data):
        if not data: data = self._data
        match_mode, re_mode, group_index = self.kwargs.get('match_mode'), self.kwargs.get('re_mode'), self.kwargs.get(
            'group_index')
        if match_mode == 'S': match_mode = re.S

        start = time.time()
        for key, re_rule in self.rules_to_dict().items():
            if re_mode == 'search':
                result = re.search(re_rule, data) if not match_mode else re.search(re_rule, data, match_mode)
                if not group_index:
                    self._result[key] = result.group() if result else ''
                else:
                    self._result[key] = result.group(group_index) if result else ''
            elif re_mode == 'match':
                result = re.match(re_rule, data) if not match_mode else re.match(re_rule, data, match_mode)
                self._result[key] = result.group(group_index) if result else ''
            elif re_mode == 'findall':
                result = re.findall(re_rule, data) if not match_mode else re.findall(re_rule, data, match_mode)
                self._result[key] = result
            else:
                raise Exception(f'Re Mode Error ... {re_mode}')
        end = time.time()
        if end - start > 1:
            print(f'maybe you need to change your re rule {end - start}')
        return data

    def update_result(self, key, value):
        if key in self._result.keys():
            if not isinstance(self._result.get(key), list):
                self._result[key] = [self._result.pop(key), value]
            elif not isinstance(value, list):
                self._result[key].append(value)
            else:
                self._result[key].extend(value)
        else:
            self._result[key] = value

    def replacer(self, key=None, value=None):

        for rule in self.rules:
            if isinstance(rule, (str, list)):
                for rl in rule:
                    if key and self.replace_key: key = key.replace(rl, '') if hasattr(key, 'replace') else key
                    value = value.replace(rl, '') if hasattr(value, 'replace') else value

            elif isinstance(rule, dict):
                for r_key, r_value in rule.items():
                    if key and self.replace_key: key = key.replace(r_key, r_value) if hasattr(key, 'replace') else key
                    value = value.replace(r_key, r_value) if hasattr(value, 'replace') else value
            else:
                print(f'Invalid rule type ... {type(rule)}')

        return (key, value) if key else value

    def get_rule_keys(self):
        keys = []
        for rule in self.rules:
            if isinstance(rule, str):
                keys.append(rule)
            elif isinstance(rule, list):
                keys.extend(rule)
            elif isinstance(rule, dict):
                keys.extend(rule.keys())
        return keys

    def rules_to_dict(self):
        rule_dict = {}
        if self.mode == 'update':
            if len(self.rules) == 1 and isinstance(self.rules[0], dict):
                rule_dict = self.rules[0]

            elif len(self.rules) == 2 and isinstance(self.rules[0], str) and isinstance(self.rules[1], str):
                rule_dict = dict([self.rules])

            elif len(self.rules) == 2 and isinstance(self.rules[0], list) and isinstance(self.rules[1], list):
                rule_dict = dict(list(zip(self.rules)))

            elif len(self.rules) == 2 and isinstance(self.rules[0], list) and isinstance(self.rules[1],
                                                                                         (str, int, float)):
                rule_dict = dict((_, self.rules[1]) for _ in self.rules[0])

            elif len(self.rules) == 2 and isinstance(self.rules[0], dict) and isinstance(self.rules[0], dict):
                for k, v in self.rules[0].items():
                    value = self.rules[1].get(v)
                    rule_dict[k] = value

            # list,list,dict
            elif len(self.rules) == 3 and isinstance(self.rules[0], list) and isinstance(self.rules[2], dict):
                for k, v in zip(self.rules[0], self.rules[1]):
                    value = self.rules[2].get(v)
                    rule_dict[k] = value
            else:
                raise Exception(f'Args Format Error ... unsupported args format : {self.rules}')

        elif self.mode == 'search':
            for rule in self.rules:
                if isinstance(rule, str):
                    rule_dict[rule] = rule
                elif isinstance(rule, list):
                    rule_dict = {**rule_dict, **dict(list(zip(rule, rule)))}
                elif isinstance(rule, dict):
                    rule_dict = {**rule_dict, **rule}
                else:
                    print(f'Rule Type Error ... {type(rule)}')

        else:
            raise Exception(f'Rule To Dict Error ... unsupported mode: {self.mode}')

        return rule_dict


class Printer(object):
    def __init__(self, data=None):
        self.data = data if data else ''
        self.data_groups = []
        self.information_extractor()
        self.printer()

    def printer(self):
        for dg in self.data_groups:
            lines = self.parse_data_group(dg)
            for line in lines:
                print(line)

    def information_extractor(self):
        if isinstance(self.data, requests.models.Response):
            data_dict = {
                'url': self.data.url,
                'body': self.data.request.body,
                'headers': self.data.request.headers,
                'cookies': self.data.request._cookies,
            }
            self.add_to_data_group('request', data_dict)

            data_dict = {
                'headers': self.data.headers,
                'cookies': self.data.cookies,
            }
            self.add_to_data_group('response', data_dict)

    def add_to_data_group(self, name, data_dict):
        group = DataGroup(name)
        for key, value in data_dict.items():
            if not value: value = ''
            if isinstance(value, str):
                group.add_info(key, value)
            else:
                group.add_data(key, value)
        self.data_groups.append(group)

    @staticmethod
    def parse_data_group(data_group):
        bar_length = data_group.bar_length
        head = f'{data_group.name} {"+" * (bar_length + len(data_group.name))}'
        info = data_group.info
        data_pool = data_group.data_pool
        lines = [head]

        # add info to lines
        for info_k, info_v in info.items():
            fmt = '{:<%d} | {:<%d}' % tuple(data_group.max_info_length)
            lines.append(fmt.format(info_k, info_v))

        lines.append('')
        # add data to lines
        for key, value in data_pool.items():
            lines.append(f'{key} {"-" * bar_length}')
            for d_key, d_value in value.items():
                fmt = '{:<%d} | {:<%d}' % tuple(data_group.max_data_length)
                lines.append(fmt.format(str(d_key), str(d_value)))

            lines.append('')
        return lines


class DataGroup(object):
    def __init__(self, name):
        self.name = name.upper()
        self.data_pool = {}
        self.info = {}
        self.max_info_length = [0, 0]
        self.max_data_length = [0, 0]
        self.bar_length = 0

    def add_data(self, title, data):
        if not data: return
        max_key_length = max([len(str(_)) for _ in data.keys()])
        max_value_length = max([len(str(_)) for _ in data.values()])

        self.max_data_length[0] = max_key_length if max_key_length > self.max_data_length[0] else self.max_data_length[
            0]
        self.max_data_length[1] = max_value_length if max_value_length > self.max_data_length[1] else \
            self.max_data_length[1]
        self.update_bar()

        self.data_pool[title] = data

    def add_info(self, title, info):
        self.max_info_length[0] = len(title) if len(title) > self.max_info_length[0] else self.max_info_length[0]
        self.max_info_length[1] = len(info) if len(info) > self.max_info_length[1] else self.max_info_length[1]
        self.info[title] = info
        self.update_bar()

    def update_bar(self):
        info_length = self.max_info_length[0] + self.max_info_length[1]
        data_length = self.max_data_length[0] + self.max_data_length[1]
        self.bar_length = info_length if info_length > data_length else data_length


class DictFactory(object):

    def del_dict_depth(self, data_dict):
        data = {}
        list_data = {}
        for key, value in data_dict.items():
            value = str(value)
            if isinstance(value, dict):
                data = {**data, **self.del_dict_depth(value)}
            elif isinstance(value, list):
                if not value:
                    continue
                if isinstance(value[0], dict):
                    list_data[key] = self.list_to_dict(value)
                else:
                    value = '|'.join(value)
                data = {**data, key: value}
            else:
                data = {**data, key: value}

        return {**data, **list_data}

    @staticmethod
    def list_to_dict(data_list):
        result = {}
        for data in data_list:
            for key, value in data.items():
                if key in [_ for _ in result.keys()]:
                    value_ = result.get(key)
                    result[key] = f'{value_}|{value}'
                else:
                    result = {**result, key: value}

        return result

    @staticmethod
    def find(target, dict_data, find_key=False):
        queue = [dict_data]
        while len(queue) > 0:
            data = queue.pop()
            for key, value in data.items():
                if key == target:
                    return value
                elif value == target and find_key:
                    return key
                elif type(value) == dict:
                    queue.append(value)
        return ''

    @staticmethod
    def print_table(data, title=None, no_print=None):
        if isinstance(data, requests.cookies.RequestsCookieJar):
            for cookie in iter(data):
                # FIXME ...
                pass

        if isinstance(data, dict):
            key_max_length = max([len(_) for _ in data.keys()])
            value_max_length = max([len(_) for _ in data.values()])
            title_index = (key_max_length + value_max_length) // 2
            if not title: title = 'dict to table'

            head = '-' * title_index + title.center(len(title) + 2, ' ') + '-' * title_index
            lines = [head]
            for key, value in data.items():
                formatter = "{:<%d} | {:<}" % key_max_length
                line = formatter.format(key, str(value))
                lines.append(line)

            # tail = '-' * title_index + '-' * len(title) + '-' * title_index + '--'
            # lines.append(tail)
            if no_print:
                return [_ + '\n' for _ in lines]
            else:
                for line in lines:
                    print(line)


def replacer(*rules, data=None, replace_key=False, mode='replace', **kwargs):
    diter = DataIter(*rules, data=data, replace_key=replace_key, mode=mode, **kwargs)
    return diter.result


def printer(data):
    return Printer(data)



import json
import logging

import urllib3
from parsel import Selector
import requests
from requests import Request, Session
from requests.cookies import merge_cookies, cookiejar_from_dict
from Dtautils.data_factory import Printer, DataGroup, DataIter
from queue import Queue

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

logger = logging.getLogger(__name__)


class SpiderUpdater(object):
    def __init__(self, url=None, body=None, header=None, cookie=None, overwrite=True):

        if not header: header = {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36'}

        self.method = 'POST' if body else 'GET'
        self.post_type = None
        if body: self.post_type = 'form' if '=' in body else 'payload'

        self._path_dict = self._string_to_dict(url, tag='path')
        self._body_dict = self._string_to_dict(body, tag='body')
        self._param_dict = self._string_to_dict(url, tag='param')
        self._header_dict = self._string_to_dict(header, tag='header')
        self.overwrite = overwrite

        cookie_dict = self._string_to_dict(cookie, tag='cookie')
        self._cookie_jar = cookiejar_from_dict(cookie_dict) if cookie else requests.cookies.RequestsCookieJar()

    @property
    def url(self):
        return self.path + self.param if self.path else ''

    @url.setter
    def url(self, url):
        self._path_dict = self._string_to_dict(url, tag='path')
        self._param_dict = self._string_to_dict(url, tag='param')

        if self.headers.get('Host'): self.update('Host', self._path_dict.get('domain'), tag='header')

    @property
    def path(self):
        assert self._path_dict, 'Please set a url for Spider'

        protocol = self._path_dict.get('protocol')
        domain = self._path_dict.get('domain')
        sub_path = '/'.join([value for key, value in self._path_dict.items() if key not in ['protocol', 'domain']])
        return f'{protocol}://{domain}/{sub_path}'

    @property
    def path_dict(self):
        return self._path_dict

    @path.setter
    def path(self, path):
        self._path_dict = self._string_to_dict(path, tag='path')

    @property
    def param(self):
        if not self._param_dict: return ''
        param_str = '&'.join([f'{key}={value}' for key, value in self._param_dict.items()])
        return '?' + param_str

    @property
    def param_dict(self):
        return self._param_dict

    @param.setter
    def param(self, params):
        self._param_dict = self._string_to_dict(params, tag='param')

    @property
    def body(self):
        if not self.post_type: return ''

        if self.post_type == 'form':
            return '&'.join([f'{key}={value}' for key, value in self._body_dict.items()])
        else:
            return json.dumps(self._body_dict)

    @property
    def body_dict(self):
        return self._body_dict

    @body.setter
    def body(self, body):
        self._body_dict = self._string_to_dict(body, tag='body')

    @property
    def headers(self):
        return self._header_dict

    @headers.setter
    def headers(self, header):
        self._header_dict = self._string_to_dict(header, tag='header')

    @property
    def cookies(self):
        return self._cookie_jar.get_dict()

    @cookies.setter
    def cookies(self, cookie):
        cookie_dict = self._string_to_dict(cookie, tag='cookie')
        self._cookie_jar = cookiejar_from_dict(cookie_dict=cookie_dict, cookiejar=self._cookie_jar,
                                               overwrite=self.overwrite)

    @property
    def cookie_jar(self):
        return self._cookie_jar

    @property
    def _key_dict(self):
        return {
            'path': list(self._path_dict.keys()),
            'param': list(self._param_dict.keys()),
            'body': list(self._body_dict.keys()),
            'header': list(self._header_dict.keys()),
            'cookie': list(self.cookies.keys()),
        }

    def update_cookie_from_header(self):
        cookie = self.headers.get('Cookie')
        if cookie:
            cookie_dict = self._string_to_dict(cookie, tag='cookie')
            self.update(cookie_dict, tag='cookie')

    def update(self, *args, tag=None, prepare=False):
        """ update tag use args

        tag: param, body, header, cookie
        args: (dict,) or (str,str) or (list,list) or (dict,dict) or (list,list,dict)
        """

        if len(args) == 1 and isinstance(args[0], dict):
            for key, value in args[0].items():
                self._update(key, value, tag=tag)

        elif len(args) == 1 and isinstance(args[0], requests.models.Response):
            self._cookie_jar.update(args[0].cookies)

        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], str):
            key, value = args
            self._update(key, value, tag=tag)

        elif len(args) == 2 and isinstance(args[0], list) and isinstance(args[1], list):
            for key, value in zip(args):
                self._update(key, value, tag=tag)

        elif len(args) == 2 and isinstance(args[0], dict) and isinstance(args[1], dict):
            for key, value in args[0].items():
                value = args[1].get(value)
                self._update(key, value, tag=tag)

        elif len(args) == 3 and isinstance(args[0], list) and isinstance(args[1], list) and isinstance(args[2], dict):
            for key, value in zip(args[0], args[1]):
                value = args[2].get(value)
                self._update(key, value, tag=tag)
        else:
            raise Exception(f'Update Error ... unsupported update args {args}')

        if prepare: return Request(url=self.url, data=self.body, headers=self.headers, cookies=self.cookies,
                                   method=self.method).prepare()

    def _update(self, key, value, tag=None):
        if not tag: tag = self._get_tag(key)

        if tag == 'param':
            self._param_dict[key] = value
        elif tag == 'body':
            self._body_dict[key] = value
        elif tag == 'header':
            self._header_dict[key] = value
        elif tag == 'cookie':
            self._cookie_jar.set(key, value)
        elif tag == 'path':
            self._path_dict[key] = value
        else:
            print(f'Update faild ... no tag {key}:{value}')

    def _get_tag(self, key):
        tag_name_list = [tag_name for tag_name, key_list in self._key_dict.items() if key in key_list]
        assert len(tag_name_list) <= 1, f'Please set tag for update, there are many tag name: {tag_name_list}'

        if tag_name_list:
            tag = tag_name_list[0]
        else:
            tag = 'param' if self.method == 'GET' else 'body'
        return tag

    def _string_to_dict(self, string, tag=None):
        result = {}
        if not string: return result
        if isinstance(string, dict): return string

        string = string.strip()

        if tag == 'path' and '://' in string:
            if '?' in string: string = string.split('?')[0]

            protocol = string.split('://', 1)[0]
            domain = string.split('://', 1)[1].split('/')[0]
            path_list = string.split('://', 1)[1].split('/')[1:]

            result['protocol'] = protocol
            result['domain'] = domain
            result = {**result, **dict(zip([str(i + 1) for i in range(len(path_list))], path_list))}

        if tag == 'param':
            if self.body: return result

            # sting : 'https://...?...'
            if '?' in string:
                if '=' in string:
                    string = string.split('?')[-1]
                    result = dict(
                        [_.split('=', 1) for _ in string.split('&') if '=' in _ and not _.endswith('=')])
                    result = {**result, **dict([(_.strip('='), '') for _ in string.split('&')
                                                if _.endswith('=') or '=' not in _])}
                else:
                    result = dict([(string, string) for _ in string.split('&')])

            elif '=' in string:
                result = dict([_.split('=', 1) for _ in string.split('&')])
            else:
                result = {}

        if tag == 'body':
            if '=' in string and '&' in string:
                result = dict([_.split('=', 1) for _ in string.split('&')])
            elif '=' in string:
                result = dict([string.split('=')])
            elif ':' in string:
                result = json.loads(string)

        if tag == 'header' or tag == 'cookie':
            split_params = ['\n', ':'] if tag == 'header' else [';', '=']
            for field in string.split(split_params[0]):
                keys = result.keys()

                key, value = field.split(split_params[1], 1)
                key, value = [key.strip(), value.strip()]

                value_in_dict = result.get(key)
                if key in keys and isinstance(value_in_dict, str):
                    result[key] = [value_in_dict, value]
                elif isinstance(value_in_dict, list):
                    result[key].append(value)
                else:
                    result[key] = value

            if tag == 'header': self._header_dict = result
            if tag == 'cookie': self._cookie_jar = cookiejar_from_dict(result)

        return result

    def preview(self, tag=None):
        d_group = DataGroup('Spider')
        d_group.add_info('url', self.url)
        d_group.add_info('param', self.param)
        d_group.add_info('body', self.body)

        preview_dict = {'headers': self.headers, 'cookies': self.cookies}
        preview_list = ['headers', 'cookies'] if not tag else [tag]

        for key, value in preview_dict.items():
            if key in preview_list:
                d_group.add_data(key, value)
            else:
                continue

        printer = Printer()
        return printer.parse_data_group(d_group)

    def __repr__(self):
        lines = self.preview()
        return '\n'.join(lines)


class SpiderExtractor(object):
    def find(self, *rules, data=None, match_mode=None, re_mode='search', group_index=None):
        result = DataIter(*rules, data=data, mode='search', match_mode=match_mode, re_mode=re_mode,
                          group_index=group_index)
        return result.result

    def css(self, *rules, data=None, extract=True, first=True, extract_key=False):

        css_tree = Selector(data)
        result = {}

        if len(rules) == 1 and isinstance(rules[0], str):
            result = self._get_css_result(css_tree, rules[0])

        elif len(rules) == 1 and isinstance(rules[0], dict):
            for key, css_rule in rules[0].items():
                if extract_key:
                    key = self._get_css_result(css_tree, key)
                    key = key.extract() if not first else key.extract_first()
                result[key] = self._get_css_result(css_tree, css_rule)
        else:
            print(f'Rule Format Error ... unsupported rule format {rules}')

        if extract and isinstance(result, dict):
            for key, value in result.items():
                result[key] = self._extract_selector(value, first=first)
        elif extract:
            result = self._extract_selector(result, first=first)

        return result

    @staticmethod
    def _extract_selector(value, first=None):
        if hasattr(value, 'extract'):
            return value.extract() if not first else value.extract_first()
        else:
            value = [_.extract() if not first else _.extract_first() for _ in value if _.extract()]
            return value[0] if len(value) == 1 else value

    @staticmethod
    def _get_css_result(css_tree, css_rule):
        if '|' not in css_rule:
            result = css_tree.css(css_rule)
        else:
            result = [css_tree.css(_) for _ in css_rule.split('|')]
        return result


class SpiderDownloader(object):
    def __init__(self, timeout=10, stream=False, verify=None, allow_redirects=True, proxies=None, hooks=None,
                 cert=None):
        self.download_count = 0
        self.session = Session()
        self._prepare_request_queue = Queue()
        self._resp_queue = Queue()

        self.timeout = timeout
        self.stream = stream
        self.verify = verify
        self.allow_redirects = allow_redirects
        self.proxies = proxies
        self.hooks = hooks
        self.cert = cert

    @property
    def prepare_request(self):
        return self._prepare_request_queue.get() if not self._prepare_request_queue.empty() else None

    @prepare_request.setter
    def prepare_request(self, req):
        self._prepare_request_queue.put(req)

    @property
    def resp(self):
        if self._resp_queue.empty(): self.request()
        return self._resp_queue.get() if not self._resp_queue.empty() else None

    @resp.setter
    def resp(self, resp):
        self._resp_queue.put(resp)

    @property
    def count(self):
        return {'req': self._prepare_request_queue.qsize(), 'resp': self._resp_queue.qsize(),
                'download': self.download_count}

    def request(self, **kwargs):
        prepared_request = self.prepare_request
        if prepared_request:
            kwargs = {'timeout': self.timeout,
                      'stream': self.stream,
                      'verify': self.verify,
                      'allow_redirects': self.allow_redirects,
                      'proxies': self.proxies,
                      'cert': self.cert, **kwargs}
            resp = self.session.send(prepared_request, **kwargs)
            self._resp_queue.put(resp)
            self.download_count += 1
            return resp
        else:
            print('No prepared request error ... spider prepare request queue is empty!')


class Spider(SpiderUpdater, SpiderDownloader, SpiderExtractor):
    def __init__(self, url=None, body=None, header=None, cookie=None, overwrite=True, timeout=10, stream=False,
                 verify=None, allow_redirects=True, proxies=None, cert=None):

        super(Spider, self).__init__(url=url, body=body, header=header, cookie=cookie, overwrite=overwrite)

        SpiderDownloader.__init__(self, timeout=timeout, stream=stream, verify=verify, allow_redirects=allow_redirects,
                                  proxies=proxies, cert=cert)

        prepare_request = Request(url=self.url, data=self.body, headers=self.headers, cookies=self.cookies,
                                  method=self.method).prepare()
        self._prepare_request_queue.put(prepare_request)

    def update(self, *args, tag=None, prepare=False):
        prepared_request = SpiderUpdater.update(self, *args, tag=tag, prepare=prepare)
        if prepare:
            self._prepare_request_queue.put(prepared_request)

    def find(self, *rules, data=None, match_mode=None, re_mode='search', group_index=None):
        if not data:
            resp = self.resp
            data = resp.text
            if data:
                if '<html' not in data and '</html>' not in data: data = json.loads(data)
            else:
                print(f'response body is empty!\n{resp}')

        return SpiderExtractor.find(self, *rules, data=data, match_mode=match_mode, re_mode=re_mode,
                                    group_index=group_index)

    def css(self, *rules, data=None, extract=True, first=True, extract_key=False):
        if not data: data = self.resp.text
        return SpiderExtractor.css(self, *rules, data=data, extract=extract, first=first, extract_key=extract_key)
